// DE_method.h: interface for the CDE_method class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DE_METHOD_H__002200BD_9F8C_4A59_B103_2261C20E80AE__INCLUDED_)
#define AFX_DE_METHOD_H__002200BD_9F8C_4A59_B103_2261C20E80AE__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "./EACommonClass/Individual.h"
#include "./EACommonClass/Rand.h"
#include "./EACommonClass/ProblemDef.h"

#define out_internal	25						// Ƶ

typedef struct strct_rewards {
	double	reward[500];
	long	size;
}strct_reward;

class CDE_method  
{
public:
	CDE_method();
	virtual ~CDE_method();

	/* ------------------------class methods------------------------ */
public:
	// public methods
	/* initialize the decision variables of different test problems */
	void init_variables();							// initialize the decision variables
	void init_normal_variables();					// initialize the normal functions' variables

	/* initialize the population */
	void init_pop();								// initialize the population
	void init_pop_random();							// initialize the population randomly

	/* evaluate the individual and population */
	void evaluate_ind(CIndividual &);				// evaluate the individual
	void evaluate_normal_ind(CIndividual &);		// evaluate the individual of normal functions

	void evaluate_pop(CIndividual *, int);			// evaluate the population

	/* execute the algorithms */
	void Run_Optimizer(int, double, int, int, int, int);		// execute the optimizer

	/* compare the two individuals */
	int  compare_ind(CIndividual *, CIndividual *);	// compare the two individuals
	void find_best_index(CIndividual *, int);		// find the index of the best individual in the population

	void sort_pop(CIndividual *, int);				// sort the population w.r.t fitness with ascend order

	/* some complementary methods */
	void random_index(int *array_index, int all_size, int size);

	void shell_sort_pop(CIndividual* ,int *, int);	// shell sort routine in ascending order
	void shell_sort_array(double *, int *, int);

	inline  int round(double x) { return int(x>0.0?(x+0.5):(x-0.5)); } // round function

	double distance(double *x1, double *x2, int size);
	double median(double *array, int size);			// get the median value of the specific array
	double mean(double *array, int size);
	void   statistic_pop_fitness(CIndividual *pop, int size);

	void display_result(int);						// display the results in the screen
	void report_result(int, ofstream &);			// report the results in the files
	void report_diversity(int, ofstream &);			// report the diversity in the files
	void report_parameter(int gen, ofstream &file);

	// need to modify for different algorithms
	double parameters[100][10];						// 
	void   read_parameters();						// read the parameters; only for parameter study
	
	/* ------------------------class members------------------------ */	
public:
	// public attributes
	CIndividual *parent_pop;				// the main population
	CIndividual *child_pop;					// the child population
	CIndividual best_individual;			// the current best individual
	CIndividual new_child;					// the new child generated by the genetic operator
	
	int pop_size;							// size of the main population
	int best_index;							// index of the current best individual
	int worst_index;

	int flag_precision;						// stop criterion

	long int max_NFFEs;						// maximal number of fitness function evaluations
	long int max_iterations;				// maximal number of iterations
	long int evaluations;					// current number of fitness function evaluations
	long int gen;							// current generation counter
	int MINIMIZE;

	double *upper_bounds;					// upper bounds of the decision variables
	double *lower_bounds;					// lower bounds of the decision variables
	double known_optimal;					// known optima of the problems

	/* only for statistical purpose */
	double  pop_diversity;
	double  max_interval;
	double  diversity_flag;					// in (0, 1]
	double	avg_fitness;
	double	min_fitness;
	double  max_fitness;
	double	media_fitness;
	double	std_fitness;

	/* some other class objectives */
	double rnd_seed;
	CRand  m_rnd;

	int						func_flag;			// flag of the function optimized
	CProblemDef				m_func;				// normal test functions

	int method_flag;
	int	*sort_index;							// only for sorting the population

	/* ----- Methods and attributes for DE algorithm ---- */
public:
	void   run_DE_method();

	double m_F;									// scale factor
	double m_CR;								// probability of DE crossover
	int    m_strategy;							// the strategy index

	/* ----- Methods and attributes for AOS-based DE algorithm ---- */
public:
	int		strategy_num;						// number of strategies in the strategy pool

	double	Reward[10];							// reward vector
	double	Quality[10];						// quality vector
	double	Probability[10];					// probability vector

	int		strategy_index[population_size];
	int		uniform_AOS;						// the baseline AOS-DE with uniform selection

	double	p_min;								// the minimal probability
	double	alpha;								// adaptation rate in [0,1]
	double	beta;								// parameter used in adaptive pursuit

	double	adaptive_alpha[10];					// implement the adaptive alpha for each strategy

	void	initial_AOS_parameter();			// initialize the AOS parameters
	void	update_quality();					// update the quality vector
	void	update_probability_PM();			// update the PM probability vector
	int		select_strategy_prob();				// select the strategy based on the probability

	void	run_DE_AOS();

	/* AOS-based DE with different credit assignment methods */
	int          window_size;					// the size of the window
	strct_reward reward_array[10];
	double       maxReward;						// the maximal reward, for normalization

	int		reward_type;						// 0: average; 1: extreme; 2: instantaneous
	int		normalized;							// normalize the reward or not
	
	void	insert_reward(int, double);			// insert the reward into the array
	void	get_reward();
	double	get_avg_reward(int strategyIndex);
	double	get_ext_reward(int strategyIndex);
	double	get_ins_reward(int strategyIndex);
	void	normalize_reward();
};

#endif // !defined(AFX_DE_METHOD_H__002200BD_9F8C_4A59_B103_2261C20E80AE__INCLUDED_)
